<?php
if ( ! class_exists( 'WCLSI_Check_For_New_Prods' ) ) :

	class WCLSI_Check_For_New_Prods {
		const LOAD_INTO_IMPORTER = 'load_into_importer';
		const IMPORT             = 'import';
		const IMPORT_AND_LINK    = 'import_and_link';
		const WCLSI_NEW_LS_PROD_BEHAVIOR = [
			self::LOAD_INTO_IMPORTER => 'Loaded into the Product Importer (default)',
			self::IMPORT             => 'Imported into WooCommerce, but not linked',
			self::IMPORT_AND_LINK    => 'Imported & linked with Woo',
		];

		static function init() {
			add_action( 'init', [ get_called_class(), 'schedule_check_for_new_ls_prods' ] );
			add_action( 'wclsi_import_on_auto_load', [ get_called_class(), 'wclsi_import_on_auto_load' ], 10, 1 );
			add_action( 'check_for_new_ls_prods', [ get_called_class(), 'check_for_new_ls_prods' ], 10 );
			add_action( 'admin_notices', [ get_called_class(), 'render_new_ls_prods_notification' ] );
		}

		/**
		 * Adds the "check_for_new_ls_prods" job to Action Scheduler
		 */
		static function schedule_check_for_new_ls_prods() {
			global $WCLSI;

			$is_enabled = $WCLSI->settings[ WCLSI_LS_TO_WC_AUTOLOAD ]['is_enabled'] ?? false;
			if ( 'true' !== $is_enabled ) {
				if ( false !== as_next_scheduled_action( 'check_for_new_ls_prods' ) ) {
					as_unschedule_all_actions( 'check_for_new_ls_prods' );
				}

				return;
			}

			if ( false === as_next_scheduled_action( 'check_for_new_ls_prods' ) ) {
				$interval = $WCLSI->settings[ WCLSI_LS_TO_WC_AUTOLOAD ]['interval'] ?? 60;
				as_schedule_recurring_action( strtotime( "+{$interval} seconds" ), $interval, 'check_for_new_ls_prods', [], 'wclsi' );
				wclsi_cleanup_duplicate_jobs( 'check_for_new_ls_prods' );
			}
		}

		/**
		 * Checks for new LightSpeed products since last import
		 */
		static function check_for_new_ls_prods() {
			if ( wclsi_skip_job() ) {
				return;
			}

			global $WCLSI, $WCLSI_SINGLE_LOAD_RELATIONS;

			$last_load_timestamp = get_option( WCLSI_LAST_LOAD_TIMESTAMP );

			if ( ! empty( $last_load_timestamp ) ) {
				// Go back 60 second in case we missed something
				$since         = date( DATE_ATOM, strtotime( $last_load_timestamp ) - 60 );
				$search_params = [
					'load_relations' => json_encode( $WCLSI_SINGLE_LOAD_RELATIONS ),
					'createTime'     => '>,' . $since,
				];

				$response = $WCLSI->make_api_call(
					"Account/$WCLSI->ls_account_id/Item/",
					'Read',
					$search_params
				);

				if ( ! is_wp_error( $response ) && $response->{'@attributes'}->count > 0 ) {
					if ( is_object( $response->Item ) ) {
						$response->Item = [ $response->Item ];
					}

					$ls_api_items         = $response->Item;
					$prod_list            = '';
					$wclsi_ids_for_import = [];

					wp_cache_add( 'wclsi_adding_new_products', true, WCLSI_CACHE );

					foreach ( $ls_api_items as $ls_api_item ) {

						// Skip over already inserted items - this could be because of the overlap
						// of going back 60 seconds with the createTime timestamp
						if ( WCLSI_Lightspeed_Prod::item_exists( $ls_api_item ) ) {
							continue;
						}

						$result     = self::add_ls_api_item( $ls_api_item );
						$wclsi_item = new WCLSI_Lightspeed_Prod( $result['wclsi_item_id'] );
						$sku        = wclsi_get_ls_sku( $wclsi_item );
						$prod_list .= "<li>Description: \"$wclsi_item->description\" SKU: \"$sku\"</li>";

						// Accumulate IDs for possible import
						$wclsi_ids_for_import[] = $result['wclsi_item_id'];
						$wclsi_ids_for_import[] = $result['wclsi_matrix_id'];
					}

					// If the import_on_auto_load settings is enabled, schedule the imports
					$autoload_after_action = $WCLSI->settings[ WCLSI_LS_TO_WC_AUTOLOAD ][ WCLSI_AUTOLOAD_AFTER_ACTION ] ?? self::LOAD_INTO_IMPORTER;
					if ( $autoload_after_action == self::IMPORT_AND_LINK ||
						$autoload_after_action == self::IMPORT ) {
						$wclsi_ids_for_import = array_unique( array_filter( $wclsi_ids_for_import ) );

						if ( ! empty( $wclsi_ids_for_import ) ) {
							foreach ( $wclsi_ids_for_import as $wclsi_id ) {
								$wclsi_item = new WCLSI_Lightspeed_Prod( $wclsi_id );
								self::schedule_import_on_autoload( $wclsi_item );
							}
						}
					}

					if ( ! empty( $prod_list ) ) {
						self::render_add_new_prod_html( $response->{'@attributes'}->count, $prod_list );
					}

					wp_cache_delete( 'wclsi_adding_new_products', WCLSI_CACHE );
				}
			}

			// Always update the WCLSI_LAST_LOAD_TIMESTAMP regardless of what we are doing
			update_option( WCLSI_LAST_LOAD_TIMESTAMP, date( DATE_ATOM, strtotime( 'now' ) ) );
		}

		/**
		 * @param WCLSI_Lightspeed_Prod $wclsi_item
		 */
		static function schedule_import_on_autoload( WCLSI_Lightspeed_Prod $wclsi_item ) {
			if ( ! $wclsi_item->is_variation_product() ) {
				as_schedule_single_action( time(), 'wclsi_import_on_auto_load', [ $wclsi_item->id ], 'wclsi' );
			}
		}

		/**
		 * @param $wclsi_item_id
		 */
		static function wclsi_import_on_auto_load( $wclsi_item_id ) {
			global $WCLSI, $WCLSI_PRODS;

			$wclsi_item = new WCLSI_Lightspeed_Prod( $wclsi_item_id );
			if ( $wclsi_item->id > 0 ) {

				$option = $WCLSI->settings[ WCLSI_LS_TO_WC_AUTOLOAD ][ WCLSI_AUTOLOAD_AFTER_ACTION ] ?? self::LOAD_INTO_IMPORTER;
				if ( self::IMPORT_AND_LINK === $option ) {
					$sync = true;
				} elseif ( self::IMPORT === $option ) {
					$sync = false;
				}

				$WCLSI_PRODS->import_item( $wclsi_item, $sync );
			}
		}

		/**
		 * @param $prod_count
		 * @param $prod_list
		 */
		private static function render_add_new_prod_html( $prod_count, $prod_list ) {
			$matrix_note =
				__(
					'<p>Note: variation products will be consolidated into their parent matrix/variable product.</p>',
					'woocommerce-lightspeed-pos'
				);

			$notification_msg =
				__(
					'Good news! %1$d new Lightspeed item(s) have been added to the Lightspeed import table:<br/>%2$s%3$s',
					'woocommerce-lightspeed-pos'
				);

			$wclsi_notifications                    = get_option( WCLSI_NOTIFICATIONS, [] );
			$wclsi_notifications['wclsi_new_items'] =
				[
					'msg'  => sprintf(
						$notification_msg,
						$prod_count,
						"<ul style='margin-left:30px;list-style:square'>$prod_list</ul>",
						$matrix_note
					),
					'type' => 'success',
				];

			update_option( WCLSI_NOTIFICATIONS, $wclsi_notifications );
		}

		static function render_new_ls_prods_notification() {
			if ( is_admin() ) {
				$wclsi_notifications = get_option( WCLSI_NOTIFICATIONS, [] );

				$notification = null;
				if ( array_key_exists( 'wclsi_new_items', $wclsi_notifications ) ) {
					$notification = $wclsi_notifications['wclsi_new_items'];
				}

				if ( ! empty( $notification ) ) {
					$class = "notice notice-{$notification[ 'type' ]} is-dismissible";
					printf( '<div class="%1$s"><p>%2$s</p></div>', $class, $notification['msg'] ); // phpcs:ignore

					unset( $wclsi_notifications['wclsi_new_items'] );
					update_option( WCLSI_NOTIFICATIONS, $wclsi_notifications );
				}
			}
		}

		/**
		 * Inserts newly-added Lightspeed products
		 */
		static function add_ls_api_item( $ls_api_item ) {
			if ( empty( $ls_api_item ) ) {
				return 0;
			}

			global $WCLSI;
			$wclsi_matrix_id = null;

			// If the product is a variation and its parent does not exist, insert its parent as well
			if ( (int) $ls_api_item->itemMatrixID > 0 &&
				empty( WCLSI_Lightspeed_Prod::get_mysql_id( null, $ls_api_item->itemMatrixID ) ) ) {
				global $WCLSI_MATRIX_LOAD_RELATIONS;

				sleep( 1 );

				$wclsi_matrix_api_item = $WCLSI->make_api_call(
					"Account/$WCLSI->ls_account_id/ItemMatrix/$ls_api_item->itemMatrixID",
					'Read',
					[ 'load_relations' => json_encode( $WCLSI_MATRIX_LOAD_RELATIONS ) ]
				);

				if ( isset( $wclsi_matrix_api_item->ItemMatrix ) ) {
					$wclsi_matrix_id = WCLSI_Lightspeed_Prod::insert_ls_api_item( $wclsi_matrix_api_item->ItemMatrix );
				}
			}

			// Check if item already exists
			$wclsi_id = WCLSI_Lightspeed_Prod::get_mysql_id( $ls_api_item->itemID, $ls_api_item->itemMatrixID );
			if ( empty( $wclsi_id ) ) {
				$wclsi_id = WCLSI_Lightspeed_Prod::insert_ls_api_item( $ls_api_item );
			}

			if ( $ls_api_item->itemMatrixID > 0 ) {
				$wclsi_parent_prod = new WCLSI_Lightspeed_Prod();
				$wclsi_parent_prod->init_via_item_matrix_id( $ls_api_item->itemMatrixID );

				// If the parent has been imported, then import the variation as well
				if ( $wclsi_parent_prod->wc_prod_id > 0 ) {
					global $WCLSI_PRODS;
					$parent_wc_prod = wc_get_product( $wclsi_parent_prod->wc_prod_id );
					$wclsi_prod     = new WCLSI_Lightspeed_Prod( $wclsi_id );
					if ( ! empty( $parent_wc_prod ) ) {
						$WCLSI_PRODS->update_matrix_variations( [ $wclsi_prod ], $parent_wc_prod );
					}
				}
			}

			return [
				'wclsi_item_id'   => $wclsi_id,
				'wclsi_matrix_id' => $wclsi_matrix_id,
			];
		}
	}

	WCLSI_Check_For_New_Prods::init();
endif;
