<?php
if ( ! class_exists( 'WCLSI_Item_Poller' ) ) :
	class WCLSI_Item_Poller {

		const LS_ITEM_TYPES = [ 'ItemMatrix', 'Item' ];
		const UPDATE_TYPES  = [
			'timeStamp'           => 'product updates',
			'ItemShops.timeStamp' => 'inventory updates',
			'Images.timeStamp'    => 'image updates',
		];

		public static function init() {
			add_action( 'init', [ get_called_class(), 'schedule_wclsi_poller' ], 10 );
			add_action( 'wclsi_poll', [ get_called_class(), 'poll' ], 10 );
			add_action( 'wclsi_update_item', [ get_called_class(), 'update_item' ], 10, 1 );
		}

		static function schedule_wclsi_poller() {
			global $WCLSI;

			$poller_enabled = $WCLSI->settings[ WCLSI_POLLER_SETTING ]['is_enabled'] ?? false;
			if ( 'true' !== $poller_enabled ) {
				if ( false !== as_next_scheduled_action( 'wclsi_poll' ) ) {
					as_unschedule_all_actions( 'wclsi_poll' );
				}

				return;
			}

			if ( false === as_next_scheduled_action( 'wclsi_poll' ) ) {
				$interval = $WCLSI->settings[ WCLSI_POLLER_SETTING ]['interval'] ?? 60;
				as_schedule_recurring_action( strtotime( "+{$interval} seconds" ), $interval, 'wclsi_poll', [], 'wclsi' );
				wclsi_cleanup_duplicate_jobs( 'wclsi_poll' );
			}
		}

		static function poll( WCLSI_Settings $ls_api = null, LSI_Import_Products $updater = null ) {
			// Limit poller to once per request
			if ( ! wp_cache_get( WCLSI_ITEM_POLL, WCLSI_CACHE ) ) {
				wp_cache_add( WCLSI_ITEM_POLL, true, WCLSI_CACHE );
			} else {
				return;
			}

			global $WCLSI, $WCLSI_PRODS;
			$updater = $updater ?: $WCLSI_PRODS;
			$ls_api  = $ls_api ?: $WCLSI;

			if ( wclsi_skip_job() || WCLSI_Lightspeed_Prod::get_item_count() == 0 ) {
				return;
			}

			$last_check = strtotime( get_option( WCLSI_LAST_SYNC_TIMESTAMP ) );
			if ( false == $last_check || self::exceeds_max_window( $last_check ) ) {
				$polling_interval = $WCLSI->settings[ WCLSI_POLLER_SETTING ]['interval'] ?? 600; // default to 10 minutes if no interval found
				$since            = strtotime( '-' . $polling_interval . ' seconds' );
			} else {
				$since = $last_check;
			}

			$matrix_items_to_update = [];
			$simple_items_to_update = [];

			foreach ( self::LS_ITEM_TYPES as $LS_ITEM_TYPE ) {
				foreach ( self::UPDATE_TYPES as $TIMESTAMP_FIELD => $UPDATE_TYPE ) {

					// Skip inventory updates for matrix products in LS
					if ( $LS_ITEM_TYPE == 'ItemMatrix' && $TIMESTAMP_FIELD == 'ItemShops.timeStamp' ) {
						continue;
					}

					$poll_results = self::poll_api( $since, $LS_ITEM_TYPE, $TIMESTAMP_FIELD, $ls_api );
					if ( ! empty( $poll_results ) ) {
						foreach ( $poll_results as $ls_item ) {
							if ( 'Item' == $LS_ITEM_TYPE ) {
								$simple_items_to_update[ (int) $ls_item->itemID ] = $ls_item;
							}

							if ( 'ItemMatrix' == $LS_ITEM_TYPE ) {
								$matrix_items_to_update[ (int) $ls_item->itemMatrixID ] = $ls_item;
							}
						}
					}
				}
			}

			$all_items_to_update = array_merge(
				array_values( $simple_items_to_update ),
				array_values( $matrix_items_to_update )
			);

			update_option( WCLSI_LAST_SYNC_TIMESTAMP, date( DATE_ATOM, strtotime( 'now' ) ) );

			if ( ! empty( $all_items_to_update ) ) {
				$number_of_items_to_update = count( $all_items_to_update );
				foreach ( $all_items_to_update as $ls_api_item ) {
					if ( $number_of_items_to_update > WCLSI_VARIATION_LIMIT ) {
						as_schedule_single_action( time(), 'wclsi_update_item', [
							$ls_api_item,
						], 'wclsi' );
					} else {
						self::update_item( $ls_api_item, $updater );
					}
				}
			}
		}

		static function poll_api( $since, $ls_item_type, $timestamp_field, WCLSI_Settings $ls_api = null, $limit = 100 ) {
			global $WCLSI_SINGLE_LOAD_RELATIONS, $WCLSI_MATRIX_LOAD_RELATIONS;

			$poll_matrix_items = $ls_item_type == 'ItemMatrix';

			$relations = $poll_matrix_items ? $WCLSI_MATRIX_LOAD_RELATIONS : $WCLSI_SINGLE_LOAD_RELATIONS;

			// Convert to string, go back 60 seconds in case we missed some updates or the action scheduler took a while to run
			$since = date( DATE_ATOM, $since - 60 );

			$search_params = [
				'load_relations' => json_encode( $relations ),
				$timestamp_field => ">,$since",
				'archived'       => true,
				'limit'          => $limit,
			];

			$next_page = true;
			$offset    = 0;
			$results   = [];
			while ( $next_page ) {

				$search_params['offset'] = $offset;

				$poll_result = $ls_api->make_api_call(
					"Account/$ls_api->ls_account_id/$ls_item_type",
					'Read',
					$search_params
				);

				if ( is_wp_error( $poll_result ) ) {
					self::log_failed_poll( $poll_result, $since );

					// Force poll result to be an empty array in the case of an error
					return $results;
				} elseif ( isset( $poll_result->{$ls_item_type} ) ) {
					if ( is_object( $poll_result->{$ls_item_type} ) ) {
						$results[] = $poll_result->{$ls_item_type};
					} elseif ( is_array( $poll_result->{$ls_item_type} ) ) {
						$results = array_merge( $results, $poll_result->{$ls_item_type} );
					}
				}

				$count = isset( $poll_result->{'@attributes'} ) ? (int) $poll_result->{'@attributes'}->count : 0;
				if ( $count < $offset + $limit || $count === 0 ) {
					$next_page = false;
				} else {
					$offset += $limit;
				}
			}
			self::log_successful_poll( $results, $since, $ls_item_type, $timestamp_field );

			return $results;
		}

		public static function update_item( $api_item, LSI_Import_Products $updater = null ) {
			$updater       = is_null( $updater ) ? new LSI_Import_Products() : $updater;
			$api_item      = (object) $api_item;
			$item_id       = property_exists( $api_item, 'itemID' ) ? $api_item->itemID : null;
			$item_mysql_id = WCLSI_Lightspeed_Prod::get_mysql_id( $item_id, $api_item->itemMatrixID );

			if ( ! empty( $item_mysql_id ) ) {
				$wclsi_prod = new WCLSI_Lightspeed_Prod( $item_mysql_id );
				WCLSI_Lightspeed_Prod::update_via_api_item( $api_item, $wclsi_prod );

				$wclsi_prod->reload();
				if ( $wclsi_prod->wc_prod_id > 0 ) {
					$wclsi_sync = (bool) get_post_meta( $wclsi_prod->wc_prod_id, WCLSI_SYNC_POST_META, true );
					if ( $wclsi_sync ) {
						wp_cache_add( 'wclsi_poller_update', true, WCLSI_CACHE );

						/**
						 * No need to update matrix variations as it should be covered
						 * by the single item poll function
						 */
						$updater->update_wc_prod( $wclsi_prod, false );

						wp_cache_delete( 'wclsi_poller_update', WCLSI_CACHE );
					}
				} else {
					// action to review/import unimported products from Importer Table via code snippet
					do_action( 'wclsi_poller_update', $api_item );
				}
			} else {
				// action to review/import un-added and unimported products from LS via code snippet
				do_action( 'wclsi_poller_update', $api_item );
			}
		}

		private static function log_successful_poll( $poll_result, $since, $ls_item_type, $timestamp_field ) {
			global $WCLSI;
			$inventory_shop_id = $WCLSI->settings[ WCLSI_INVENTORY_SHOP_ID ] ?? 0;

			$poll_matrix_items = $ls_item_type == 'ItemMatrix';

			$update_type = self::UPDATE_TYPES[ $timestamp_field ];

			$update_description = "Syncing $ls_item_type $update_type for: " . PHP_EOL;

			if ( count( $poll_result ) > 0 ) {
				foreach ( $poll_result as $item ) {
					if ( $poll_matrix_items ) {
						$update_description .=
							'Matrix item id: ' . $item->itemMatrixID .
							', description: ' . $item->description . PHP_EOL;
					} else {
						$inventory_log = '';
						if ( $update_type === 'inventory updates' ) {
							$inventory_log =
								', New inventory: ' . $item->ItemShops->ItemShop[ $inventory_shop_id ]->qoh ?? "Quantity not available because the item shop doesn't exist";
						}

						$update_description .=
							'Item id: ' . $item->itemID .
							', description: ' . $item->description . $inventory_log . PHP_EOL;
					}
				}

				$msg = '---- Running LS auto-poll ----' . PHP_EOL .
						'---- Looking up changes since: UTC-' . $since . ' ----' . PHP_EOL;
				WCLSI_Logger::log_sync_event( $msg, $update_description );
				WCLSI_Logger::log_debug_event( $msg, $update_description );
			} else {
				$msg = "---- Running LS auto-poll for $ls_item_type $update_type ----" . PHP_EOL .
						'---- Looking up changes since: ' . $since . ' ----' . PHP_EOL .
						'No changes found' . PHP_EOL;

				WCLSI_Logger::log_debug_event( $msg );
			}
		}

		private static function log_failed_poll( $poll_result, $since ) {
			$error_msg = '---- Automated sync failed - since timeStamp: ' . $since . ' ---';
			WCLSI_Logger::log_error( $error_msg, $poll_result );
		}

		/**
		 * @param $last_check
		 *
		 * @return bool
		 */
		private static function exceeds_max_window( $last_check ): bool {
			// max window of 24hrs
			$max_window = strtotime( 'now' ) - 864000;

			return $last_check < $max_window;
		}
	}
endif;

WCLSI_Item_Poller::init();
