<?php
/**
 * Integrations Page. Sets up various options to initialize communication with LightSpeed.
 *
 * @class LSI_Init_Settings
 */

// phpcs:disable
if ( ! class_exists( 'WCLSI_Settings' ) ) :

	class WCLSI_Settings extends WC_Integration {

		const SETTINGS_TABS = [
			[
				'label' => 'Store Settings',
				'slug'  => 'wclsi_store_settings',
			],
			[
				'label'    => 'Woo Product Sync',
				'slug'     => 'wclsi_woo_product_sync_settings',
				'data_tip' => 'Settings for when product data is pulled in from Lightspeed into WooCommerce (LS→WC).',
			],
			[
				'label'    => 'Lightspeed Product Sync',
				'slug'     => 'wclsi_ls_product_sync_settings',
				'data_tip' => 'Settings for when product data is sent from WooCommerce to Lightspeed (WC→LS).',
			],
			[
				'label'    => 'Sales Sync',
				'slug'     => 'wclsi_sales_sync_settings',
				'data_tip' => 'Settings for when sales data is sent from WooCommerce to Lightspeed (WC→LS).',
				'warning'  => 'Beta',
			],
			[
				'label' => 'Categories',
				'slug'  => 'wclsi_category_sync_settings',
			],
			[
				'label' => 'Background Jobs',
				'slug'  => 'wclsi_background_jobs',
			],
		];

		/** @var $connect_url */
		public $connect_url = null;

		/**
		 * @var mixed|void|null
		 */
		public $ls_account_id = null;

		/**
		 * @var mixed|string
		 */
		public $current_tab = 'wclsi_store_settings';

		/**
		 * @var array
		 */
		private $init_errors = [];

		/**
		 * @var mixed|string|void
		 */
		public $token = '';

		/**
		 * @var WP_MOSAPICall
		 */
		public $MOSAPI;

		public $store_timezone;

		public $store_name;

		/** @var bool */
		private static $reconnected = false;

		function __construct() {
			$this->connect_url = apply_filters( 'wclsi_connect_host', 'https://connect.kestrelwp.io' );

			$this->set_header_info();
			$this->init_settings();
			$this->check_for_lightspeed_token();
			$this->token         = get_option( 'wclsi_oauth_token' );
			$this->ls_account_id = get_option( 'wclsi_account_id' );

			$this->current_tab = $_GET['wclsi-settings-tab'] ?? $this->current_tab;

			$this->init_store_data();
			$this->init_form_fields();

			if ( ! empty( $this->ls_account_id ) && ! empty( $this->token ) ) {
				$this->MOSAPI = new WP_MOSAPICall( $this->ls_account_id, $this->token );
			}

			add_action( "woocommerce_update_options_integration_{$this->id}", [ $this, 'process_admin_options' ] );
			add_action( 'admin_notices', [ $this, 'display_init_errors' ] );
			add_filter( 'woocommerce_settings_api_sanitized_fields_' . $this->id, [ $this, 'sanitize_settings' ] );
			add_action( 'admin_enqueue_scripts', [ $this, 'add_wclsi_script' ], 9 );
			add_action( 'admin_enqueue_scripts', [ $this, 'add_wclsi_script_vars' ], 10 );
			add_action( 'wp_enqueue_scripts', [ $this, 'add_wclsi_script' ], 9 );
			add_action( 'wp_enqueue_scripts', [ $this, 'add_wclsi_script_vars' ], 10 );
			add_action( 'wp_ajax_clear_table_data_ajax', [ $this, 'clear_table_data_ajax' ] );

			// display a confirmation message when reconnecting to Lightspeed
			add_action( 'all_admin_notices', function () {

				if ( ! static::$reconnected && ! empty( $this->init_errors ) && ! empty( $_GET['wclsi_access_token'] ) && isset( $_GET['tab'], $_GET['section'], $_GET['reconnect'] ) && 'integration' === $_GET['tab'] && 'lightspeed-integration' === $_GET['section'] && 'true' === $_GET['reconnect'] ) :

					?>
					<div class="notice notice-success">
						<p><?php esc_html_e( 'Successfully reconnected to Lightspeed!', 'woocommerce-lightspeed-pos' ); ?></p>
					</div>
					<?php

					static::$reconnected = true;

				endif;
			} );
		}

		/**
		 * @return string
		 */
		function get_settings_option_key() {
			return "{$this->plugin_id}{$this->id}_settings";
		}

		/**
		 * Override this method and just return settings
		 *
		 * @param $settings
		 *
		 * @return mixed
		 * @see process_admin_options()
		 */
		function sanitize_settings( $settings ) {
			return $settings;
		}

		function admin_options() {
			$this->display_errors();
			$this->check_for_reqd_fields();
			?>
			<h3><?php echo $this->method_title; ?></h3>
			<?php echo wpautop( $this->method_description ); ?>
			<?php $this->get_setting_tabs(); ?>
			<?php $this->generate_settings_info_html(); ?>
			<table <?php echo $this->get_settings_table_attrs(); ?>>
				<?php $this->generate_settings_table_headers_html(); ?>
				<?php $this->generate_settings_html(); ?>
			</table>
			<input type="hidden" name="section" value="<?php echo $this->id; ?>"/>
			<?php
		}

		function get_settings_table_attrs(
			$attrs = [
				'class' => 'form-table',
				'id'    => 'wclsi-settings-table',
			]
		): string {
			if ( $this->current_tab == 'wclsi_background_jobs' ) {
				$attrs['class'] .= ' wc_gateways widefat';
			}

			$attr_html = '';
			foreach ( $attrs as $attr_key => $attr_value ) {
				$attr_html .= "$attr_key=\"$attr_value\" ";
			}

			return $attr_html;
		}

		function generate_settings_info_html() {
			if ( $this->current_tab === 'wclsi_store_settings ' ) {
				$this->render_api_success_message();
			}

			if ( $this->current_tab === 'wclsi_background_jobs' ) {
				$action_scheduler_url = admin_url( 'tools.php?page=action-scheduler&status=pending' );
				?>
				<p class="notice notice-info" style="padding: 10px">
					<b>Note</b>: Background jobs are run using <a href="<?php echo $action_scheduler_url; ?>" target="_blank">Action
						Scheduler</a>.
				</p>
				<p class="notice notice-warning" style="padding: 10px">
					<b>Warning</b>: jobs that run on an interval of <b>less than a minute</b> may cause
					<b>high CPU usage, large log files, and a bloated database</b>.
				</p>
				<?php
			}
		}

		function generate_settings_table_headers_html() {
			if ( $this->current_tab === 'wclsi_background_jobs' ) {
				?>
				<thead>
				<tr>
					<th class="job">Job</th>
					<th class="desc">Description</th>
					<th class="enabled">Enabled</th>
					<th class="job_settings">Job Settings</th>
				</tr>
				</thead>
				<?php
			} else {
				echo '';
			}
		}

		function get_setting_tabs() {
			global $wp;
			?>
			<nav class="nav-tab-wrapper woo-nav-tab-wrapper">
				<?php
				foreach ( self::SETTINGS_TABS as $tab ) {
					$href           = esc_url( add_query_arg( [ 'wclsi-settings-tab' => $tab['slug'] ], $_SERVER['REQUEST_URI'] ) );
					$active         = $this->current_tab === $tab['slug'] ? 'nav-tab-active' : '';
					$data_tip       = array_key_exists( 'data_tip', $tab ) ? "data-tip='{$tab['data_tip']}'" : '';
					$data_tip_class = array_key_exists( 'data_tip', $tab ) ? 'help_tips tips' : '';
					$warning        = array_key_exists( 'warning', $tab ) ? $tab['warning'] : '';
					?>
					<a href="<?php echo $href; ?>"
						class="nav-tab <?php echo $active; ?> <?php echo $data_tip_class; ?>" <?php echo $data_tip; ?>>
						<div>
							<span><?php echo $tab['label']; ?></span>

							<?php
							echo $warning ? '<span style="
                                                            border-radius: 9px;
                                                            background-color: #d63638;
                                                            padding: 3%;
                                                            color: #ffffff;
                                                            font-size: 80%;
                                                            font-weight: normal"
                                                   >' . $warning . '</span>' : ''
							?>
						</div>
					</a>
				<?php } ?>
			</nav>
			<?php
		}

		function add_wclsi_script() {
			$suffix = defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ? '' : '.min';
			$deps   = is_admin() ? [ 'jquery', 'thickbox', 'jquery-tiptip' ] : [ 'jquery' ];
			wp_enqueue_script(
				'wclsi-admin-js',
				plugins_url( "assets/js/wclsi-admin{$suffix}.js", __DIR__ ),
				$deps,
				WCLSI_VERSION,
				true
			);
		}

		function add_wclsi_script_vars() {
			global $WCLSI_objectL10n;

			$script_params = [
				'PLUGIN_PREFIX_ID'  => "{$this->plugin_id}{$this->id}_",
				'SCRIPT_DEBUG'      => defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG,
				'IS_ADMIN'          => is_admin(),
				'AJAX_URL'          => admin_url( 'admin-ajax.php' ),
				'WCLSI_NONCE'       => wp_create_nonce( 'wclsi_ajax' ),
				'IMPORTER_PAGE_URL' => WCLSI_ADMIN_URL,
			];

			$wclsi_admin_script_vars = apply_filters( 'wclsi_admin_script_vars', $script_params );
			wp_localize_script( 'wclsi-admin-js', 'wclsi_admin', $wclsi_admin_script_vars );
			wp_localize_script( 'wclsi-admin-js', 'objectL10n', $WCLSI_objectL10n );
		}

		/**
		 * Will only display errors if the plugin has initialized
		 */
		function check_for_reqd_fields() {
			$wclsi_initialized = get_option( 'wclsi_initialized' );
			if ( ! $wclsi_initialized ) {
				return;
			}

			if ( empty( $this->store_timezone ) ) {
				?>
				<div class="error">
					<p><?php echo __( 'Could not find store timezone. This is required for syncing products. Please click on "Connect to Lightspeed" before starting the syncing process with Lightspeed.', 'woocommerce-lightspeed-pos' ); ?></p>
				</div>
				<?php
			}
		}

		static function get_store_settings() {
			return [
				'ls_account_id'                   => [
					'title'       => __( 'Account ID', 'woocommerce-lightspeed-pos' ),
					'type'        => 'hidden_custom',
					'desc_tip'    => true,
					'default'     => '',
					'description' => '',
				],
				'connect_to_ls'                   => [
					'title'       => __( 'Enable Lightspeed API', 'woocommerce-lightspeed-pos' ),
					'type'        => 'wclsi_ls_button_custom',
					'desc_tip'    => true,
					'default'     => '',
					'description' => __(
						'Click the button to connect your Lightspeed Retail account to WooCommerce.',
						'woocommerce-lightspeed-pos'
					),
				],
				'ls_inventory_store'              => [
					'title'             => __( 'Primary Lightspeed Inventory Store', 'woocommerce-lightspeed-pos' ),
					'type'              => 'inventory_radio_custom',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'This plugin is capable of looking up and pushing inventory to only one of your Lightspeed ' .
						'stores, please select which store you\'d like to sync inventory with.',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_import_and_link_all_prods' => [
					'title'             => __( 'Bulk Import & Link', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_import_and_link_all_prods',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Will do a bulk Import of all the Lightspeed products in the Importer and ' .
						'link them with WooCommerce.',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_delete_all_loaded_prods'   => [
					'title'             => __( 'Clear Import Table', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_delete_all_loaded_prods',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Will permanently delete all the products loaded into the import table. ' .
						'Warning: this will permanently destroy the link between WooCommerce products and their ' .
						'corresponding Lightspeed product.',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_debug_mode'                => [
					'title'             => __( 'Debug Mode', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_debug_mode',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Activate additional logging to aid in debugging',
						'woocommerce-lightspeed-pos'
					),
				],
			];
		}

		static function get_ls_product_settings() {
			return [
				'wclsi_ls_selective_sync' => [
					'title'             => __( 'Selective sync for LightSpeed products', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_ls_selective_sync',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Select which product fields should get updated for a Lightspeed product when it syncs ' .
						'data from WooCommerce.',
						'woocommerce-lightspeed-pos'
					),
				],
			];
		}

		static function get_woo_product_settings() {
			$currency_symbol = get_woocommerce_currency_symbol();

			return [
				'wclsi_import_status'              => [
					'title'             => __( 'Product status on import', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_import_status',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'The product status on import. Publish will make the product immediately visible on your ' .
						'store catalog upon import. Draft will keep the product hidden until it is published.' .
						'This setting will apply to all products on import.',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_sku_strategy'               => [
					'title'             => __( 'SKU strategy', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_sku_strategy',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						"Lightspeed has three different SKU values you can sync from: System SKU, Custom SKU, and Manufacturer's SKU. " .
						"Plugin default will prioritize Custom SKU, then Manufacturer's SKU, and lastly, System SKU.",
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_image_compression_strategy' => [
					'title'             => __( 'Image Compression strategy', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_image_compression_strategy',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'The plugin uses Cloudinary to download images from Lightspeed. The options are low, better (eco), or best. ' .
						'Plugin default will use the eco setting which results in smaller files of slightly lower visual quality.',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_handle_archived_products'   => [
					'title'             => __( 'Handle Archived Lightspeed Products on Update', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_handle_archived_products',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Choose how to handle products archived in Lightspeed when a product is updated. ' .
						'Plugin default will do nothing to Woo versions of products.',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_ignore_archived_ls_prods'   => [
					'title'             => __( 'Ignore Archived Lightspeed Products When Loading Products', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_ignore_archived_ls_prods',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'default'           => true,
					'description'       => __(
						'When checked, will ignore all archived Lightspeed products upon loading products into ' .
						'WooCommerce.',
						'woocommerce-lightspeed-pos' ),
				],
				'wclsi_empty_on_zero_sale_price'   => [
					'title'             => __( 'Set empty Sale Price on zero', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_empty_on_zero_sale_price',
					'custom_attributes' => '',
					'default'           => true,
					'desc_tip'          => true,
					'description'       => __(
						"If the Lightspeed Sale Price is {$currency_symbol}0.00, this setting will set the Sale Price " .
						'to an empty value of the corresponding WooCommerce product.',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_prune_deleted_variations'   => [
					'title'             => __( 'Prune deleted Lightspeed variations', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_prune_deleted_variations',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'default'           => true,
					'description'       => __(
						'Only works with manual updates: When checked, deletes WooCommerce variations that correspond to ones deleted in Lightspeed ' .
						'matrix products.',
						'woocommerce-lightspeed-pos' ),
				],
				'wclsi_wc_selective_sync'          => [
					'title'             => __( 'Selective sync for WooCommerce products', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_wc_selective_sync',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Select which product fields should get updated for a WooCommerce product when it syncs ' .
						'data from Lightspeed.',
						'woocommerce-lightspeed-pos'
					),
				],
			];
		}

		static function get_sales_settings() {
			return [
				'wclsi_sales_sync'       => [
					'title'             => __( 'Accept Beta Terms and Start Syncing Sales', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_sales_sync_checkbox',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'default'           => false,
					'description'       => __(
						'Accept terms of Beta release in order to begin Syncing inventory through sales. '
					),
				],
				'wclsi_sales_store'      => [
					'title'             => __( 'Store', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_sales_store',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'This is your primary inventory store',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_sales_employee'   => [
					'title'             => __( 'Employee', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_sales_employee',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Choose which employee to sync sales under',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_sales_register'   => [
					'title'             => __( 'Register', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_sales_register',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Choose which register to sync sales under',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_export_tax_rates' => [
					'title'             => __( 'Export Tax Rates to Lightspeed', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_export_tax_rates_button',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Export tax rates from WC -> LS for syncing sales',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_delete_tax_rates' => [
					'title'             => __( 'Remove Tax Rates from Lightspeed', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_delete_tax_rates_button',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Remove tax categories in Lightspeed that were generated by this plug-in',
						'woocommerce-lightspeed-pos'
					),
				],
			];
		}

		static function get_category_settings() {
			return [
				'import_categories'             => [
					'title'             => __( 'Lightspeed Categories', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_import_cats_button',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'This is a one time action that will import existing Lightspeed categories. ' .
						'Newly added Lightspeed categories will not be automatically synced! ',
						'woocommerce-lightspeed-pos'
					),
				],
				'delete_category_cache'         => [
					'title'             => __( 'Delete Cached Lightspeed Categories', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_delete_cats_button',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'This is a one time action that will delete cached Lightspeed categories stored in the WordPress database.',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_remove_default_category' => [
					'title'             => __( 'Remove default category', 'woocommerce-lightspeed-pos' ),
					'type'              => 'wclsi_remove_default_category',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Remove the WooCommerce default category from products when importing or updating from Lightspeed.',
						'woocommerce-lightspeed-pos'
					),
				],
			];
		}

		static function get_background_job_settings() {
			return [
				'wclsi_poller_setting'           => [
					'title'             => __(
						'Lightspeed Product Changes',
						'woocommerce-lightspeed-pos'
					),
					'type'              => 'wclsi_poller_setting',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Polls for Lightspeed product changes and updates the linked WooCommerce Product. ' .
						'Properties such as stock quantity, images, and descriptions will be updated.',
						'woocommerce-lightspeed-pos'
					),
				],
				'ls_to_wc_auto_load'             => [
					'title'             => __(
						'New Lightspeed Products',
						'woocommerce-lightspeed-pos'
					),
					'type'              => 'wclsi_ls_to_wc_auto_load',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Polls for new Lightspeed products. ' .
						'Products can be automatically imported & linked or loaded into the Product Importer to be imported later. ',
						'woocommerce-lightspeed-pos'
					),
				],
				'wclsi_poll_prod_attribute_sets' => [
					'title'             => __(
						'Lightspeed Product Attribute Sets',
						'woocommerce-lightspeed-pos'
					),
					'type'              => 'wclsi_poll_prod_attribute_sets',
					'custom_attributes' => '',
					'desc_tip'          => true,
					'description'       => __(
						'Polls for new Lightspeed product attribute sets. ' .
						'This will automatically create new product attributes in WooCommerce under Products -> Attributes. ',
						'woocommerce-lightspeed-pos'
					),
				],
			];
		}

		function init_form_fields() {
			switch ( $this->current_tab ) {
				case 'wclsi_store_settings':
					$this->form_fields = self::get_store_settings();
					break;
				case 'wclsi_ls_product_sync_settings':
					$this->form_fields = self::get_ls_product_settings();
					break;
				case 'wclsi_woo_product_sync_settings':
					$this->form_fields = self::get_woo_product_settings();
					break;
				case 'wclsi_sales_sync_settings':
					$this->form_fields = self::get_sales_settings();
					break;
				case 'wclsi_category_sync_settings':
					$this->form_fields = self::get_category_settings();
					break;
				case 'wclsi_background_jobs':
					$this->form_fields = self::get_background_job_settings();
					break;
				default:
					$this->form_fields = [];
			}
		}

		/*****************
		 * Form Elements *
		 *****************/
		function generate_inventory_radio_custom_html( $key, $data ) {
			$shop_data = get_option( 'wclsi_shop_data' );

			if ( false !== $shop_data && is_object( $shop_data['ls_store_data'] ) ) {
				$shop_data = $shop_data['ls_store_data'];
				$field     = "{$this->plugin_id}{$this->id}_{$key}";

				if ( is_array( $shop_data->Shop ) ) {
					$defaults = [
						'class'             => 'button-secondary',
						'css'               => '',
						'custom_attributes' => [],
						'desc_tip'          => false,
						'description'       => '',
						'title'             => '',
					];

					$data = wp_parse_args( $data, $defaults );

					ob_start();
					?>
					<tr>
						<?php echo $this->tooltip_th( $field, $data ); ?>
						<td class="forminp">
							<fieldset>
								<?php echo $this->get_description_html( $data ); ?>
								<?php foreach ( $shop_data->Shop as $store ) : ?>
									<label>
										<input
												type="radio"
												name="<?php echo $field; ?>"
												id="<?php echo $field; ?>"
												value="<?php echo $store->shopID; ?>"
											<?php
											if ( isset( $this->settings[ WCLSI_INVENTORY_SHOP_ID ] ) ) {
												if ( $this->settings[ WCLSI_INVENTORY_SHOP_ID ] == $store->shopID ) {
													echo 'checked';
												}
											}
											?>
										/>
										<?php echo $store->name; ?>
									</label>
									<br/>
								<?php endforeach; ?>
							</fieldset>
						</td>
					</tr>
					<?php
					return ob_get_clean();
				} else {
					?>
					<input
					type="hidden"
					name="<?php echo $field; ?>"
					id="name="<?php echo $field; ?>"
					value="<?php echo $this->settings[ WCLSI_INVENTORY_SHOP_ID ]; ?>"
					/>
					<?php
				}
			}

			return '';
		}

		function generate_hidden_custom_html( $key, $data ) {
			$wclsi_account_id = get_option( 'wclsi_account_id' );
			ob_start();
			if ( false !== $wclsi_account_id ) {
				?>
				<input
						type="hidden"
						id="<?php echo "{$this->plugin_id}{$this->id}_{$key}"; ?>"
						name="<?php echo "{$this->plugin_id}{$this->id}_{$key}"; ?>"
						value="<? echo $wclsi_account_id ?>"
				/>
				<?php
			}

			return ob_get_clean();
		}

		function generate_wclsi_debug_mode_html( $key, $data ) {
			return $this->single_checkbox( $key, $data );
		}

		function generate_wclsi_remove_default_category_html( $key, $data ) {
			return $this->single_checkbox( $key, $data );
		}

		function generate_wclsi_import_and_link_all_prods_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}
			$field = "{$this->plugin_id}{$this->id}_{$key}";
			$data  = wp_parse_args( $data );
			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<fieldset>
						<legend class="screen-reader-text"><span><?php echo wp_kses_post( $data['title'] ); ?></span>
						</legend>
						<?php echo $this->get_description_html( $data ); ?>
						<button class="button" type="button" id="wc-import-all-prods" style="font-size: 14px;">
							<?php echo __( 'Import and Link all the products in the Importer', 'woocommerce-lightspeed-pos' ); ?>
						</button>
					</fieldset>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_delete_all_loaded_prods_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}
			$field   = "{$this->plugin_id}{$this->id}_{$key}";
			$data    = wp_parse_args( $data );
			$caution = "<span style='color: #ff0000'>&#9888;</span> ";
			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<fieldset>
						<legend class="screen-reader-text"><span><?php echo wp_kses_post( $data['title'] ); ?></span>
						</legend>
						<?php echo $this->get_description_html( $data ); ?>
						<button class="button" type="button" id="wclsi-clear-loaded-prods" style="font-size: 14px;">
							<?php echo $caution . __( 'Clear All Loaded Products', 'woocommerce-lightspeed-pos' ); ?>
						</button>
					</fieldset>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}


		function generate_wclsi_sales_sync_checkbox_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			ob_start();
			?>
			<div class="notice notice-warning inline" style="font-size: 100%;">
				<p>Welcome to the beta release of our sales syncing feature! A few things you should know before opting
					in to sales syncing:</p>
				<ul style="list-style-type: square; padding-left: 15px">
					<li>
						This feature is a work in progress and may not work as intended. By choosing to opt in, you
						accept any and all risks associated with it.
					</li>
					<li>
						The full release of this feature and support for it may be moved to a separate extension at a
						later date.
					</li>
					<li>
						Please read the
						<a href="https://docs.google.com/document/d/1OKWS81eyEjlRactexX8ww7tJrW5F_HSw2TD7CYO6jMU/edit?usp=sharing">documentation</a>
						for sales syncing before opting in.
					</li>
					<li>
						Syncing sales with compound tax rates is not available.
					</li>
					<li>
						Syncing sales with inclusive tax rates is not available.
					</li>
					<li>
						We encourage you to reach out to our support staff with questions or suggestions regarding this
						feature! Please don’t hesitate to reach out to us with a
						<a href="https://woocommerce.com/my-account/create-a-ticket/">support ticket</a>.
					</li>
				</ul>
			</div>
			<?php
			echo $this->single_checkbox( $key, $data );

			return ob_get_clean();
		}


		function generate_wclsi_sales_store_html( $key, $data ) {

			global $WCLSI;

			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			if ( ! array_key_exists( WCLSI_SALES_SYNC, $WCLSI->settings ) ||
				'true' !== $WCLSI->settings[ WCLSI_SALES_SYNC ] ) {
				return '';
			}

			$field = "{$this->plugin_id}{$this->id}_{$key}";
			$data  = wp_parse_args( $data );

			$shop_data  = get_option( 'wclsi_shop_data' );
			$shop_data  = $shop_data['ls_store_data'];
			$store_name = '';
			if ( is_array( $shop_data->Shop ) ) {
				foreach ( $shop_data->Shop as $store ) {
					if ( $store->shopID == $this->settings[ WCLSI_INVENTORY_SHOP_ID ] ) {
						$store_name = $store->name;
					}
				}
			} else {
				$store_name = $shop_data->Shop->name;
			}

			ob_start();
			?>
			<tr style="vertical-align: top;">
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp" style="vertical-align: top; padding: 15px 10px;">
					<fieldset>
						<?php echo $this->get_description_html( $data ); ?>
						<span>
							<?php echo $store_name; ?>
						</span>
						<br/>
					</fieldset>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_sales_employee_html( $key, $data ) {
			global $WCLSI;
			global $WCLSI_SYNC_SALES;

			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			if ( ! array_key_exists( WCLSI_SALES_SYNC, $WCLSI->settings ) ||
				'true' !== $WCLSI->settings[ WCLSI_SALES_SYNC ] ) {
				return '';
			}

			$field           = "{$this->plugin_id}{$this->id}_{$key}";
			$data            = wp_parse_args( $data );
			$selected_option = isset( $this->settings[ $key ] ) ? $this->settings[ $key ] : 'plugin_default';
			$employees       = $WCLSI_SYNC_SALES->get_ls_employees();

			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<div style="max-width: 400px;">
						<fieldset>
							<legend class="screen-reader-text">
								<span><?php echo wp_kses_post( $data['title'] ); ?></span>
							</legend>
							<?php echo $this->get_description_html( $data ); ?>
						</fieldset>
						<select name="<?php echo $field; ?>" id="<?php echo $field; ?>" style="width: inherit;">
							<?php
							foreach ( $employees as $option ) {
								$selected = $selected_option === $option['value'] ? 'selected' : '';
								echo "<option value='{$option['value']}' {$selected}>{$option['text']}</option>";
							}
							?>
						</select>
					</div>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_sales_register_html( $key, $data ) {
			global $WCLSI;
			global $WCLSI_SYNC_SALES;

			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			if ( ! array_key_exists( WCLSI_SALES_SYNC, $WCLSI->settings ) ||
				'true' !== $WCLSI->settings[ WCLSI_SALES_SYNC ] ) {
				return '';
			}

			$field           = "{$this->plugin_id}{$this->id}_{$key}";
			$data            = wp_parse_args( $data );
			$selected_option = isset( $this->settings[ $key ] ) ? $this->settings[ $key ] : 'plugin_default';
			$registers       = $WCLSI_SYNC_SALES->get_ls_registers();

			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<div style="max-width: 400px;">
						<fieldset>
							<legend class="screen-reader-text">
								<span><?php echo wp_kses_post( $data['title'] ); ?></span>
							</legend>
							<?php echo $this->get_description_html( $data ); ?>
						</fieldset>
						<select name="<?php echo $field; ?>" id="<?php echo $field; ?>" style="width: inherit;">
							<?php
							foreach ( $registers as $option ) {
								$selected = $selected_option === $option['value'] ? 'selected' : '';
								echo "<option value='{$option['value']}' {$selected}>{$option['text']}</option>";
							}
							?>
						</select>
					</div>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_export_tax_rates_button_html( $key, $data ) {
			global $WCLSI;

			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			if ( ! array_key_exists( WCLSI_SALES_SYNC, $WCLSI->settings ) ||
				'true' !== $WCLSI->settings[ WCLSI_SALES_SYNC ] ) {
				return '';
			}

			$field = "{$this->plugin_id}{$this->id}_{$key}";
			$data  = wp_parse_args( $data );

			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<div style="max-width: 400px;">
						<fieldset>
							<legend class="screen-reader-text">
								<span><?php echo wp_kses_post( $data['title'] ); ?></span>
							</legend>
							<?php echo $this->get_description_html( $data ); ?>
							<button id="wclsi-export-tax-rates" type="button" class="button-secondary">Export Tax
								Rates
							</button>
							<?php wp_nonce_field( 'wclsi_ajax', 'wclsi_nonce', false ); ?>
						</fieldset>

					</div>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_delete_tax_rates_button_html( $key, $data ) {
			global $WCLSI;

			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			if ( ! array_key_exists( WCLSI_SALES_SYNC, $WCLSI->settings ) ||
				'true' !== $WCLSI->settings[ WCLSI_SALES_SYNC ] ) {
				return '';
			}

			$field = "{$this->plugin_id}{$this->id}_{$key}";
			$data  = wp_parse_args( $data );

			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<div style="max-width: 400px;">
						<fieldset>
							<legend class="screen-reader-text">
								<span><?php echo wp_kses_post( $data['title'] ); ?></span>
							</legend>
							<?php echo $this->get_description_html( $data ); ?>
							<button id="wclsi-delete-tax-rates" type="button" class="button-secondary">Delete Tax
								Rates
							</button>
							<?php wp_nonce_field( 'wclsi_ajax', 'wclsi_nonce', false ); ?>
						</fieldset>

					</div>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_delete_cats_button_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			$field = "{$this->plugin_id}{$this->id}_{$key}";
			$data  = wp_parse_args( $data );
			ob_start();

			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<fieldset>
						<legend class="screen-reader-text">
							<span><?php echo wp_kses_post( $data['title'] ); ?></span>
						</legend>
						<?php echo $this->get_description_html( $data ); ?>
						<button id="wclsi-delete-cat-cache" type="button" class="button-secondary">Delete Cache</button>
						<div id="wclsi-import-cats-progress" style="display: inline-block;"></div>
						<?php wp_nonce_field( 'wclsi_ajax', 'wclsi_nonce', false ); ?>
					</fieldset>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_import_cats_button_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			$field = "{$this->plugin_id}{$this->id}_{$key}";
			$data  = wp_parse_args( $data );
			ob_start();

			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<fieldset>
						<legend class="screen-reader-text">
							<span><?php echo wp_kses_post( $data['title'] ); ?></span>
						</legend>
						<?php echo $this->get_description_html( $data ); ?>
						<button id="wclsi-import-cats" type="button" class="button-secondary">Import Categories</button>
						<div id="wclsi-import-cats-progress" style="display: inline-block;"></div>
						<?php wp_nonce_field( 'wclsi_ajax', 'wclsi_nonce', false ); ?>
					</fieldset>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_poll_prod_attribute_sets_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			$field             = "{$this->plugin_id}{$this->id}_{$key}";
			$data              = wp_parse_args( $data );
			$is_enabled        = $this->settings[ $key ]['is_enabled'] ?? false;
			$selected_interval = $this->settings[ $key ]['interval'] ?? null;
			ob_start();
			?>
			<tr style="vertical-align: top;">
				<td scope="row" class="titledesc">
					<label for="<?php echo esc_attr( $field ); ?>"><?php echo wp_kses_post( $data['title'] ); ?></label>
				</td>
				<td>
					<?php echo $data['description']; ?>
				</td>
				<td class="forminp" style="width: 1em; vertical-align: top; padding: 15px 10px;">
					<fieldset>
						<label class="wclsi-switch">
							<input
									type="checkbox"
									id="<?php echo "{$field}[is_enabled]"; ?>"
									name="<?php echo "{$field}[is_enabled]"; ?>"
									value="true"
								<?php echo $is_enabled ? 'checked' : ''; ?>
							/>
							<span class="wclsi-slider round"></span>
						</label>
					</fieldset>
				</td>
				<td>
					<fieldset>
						<label for="<?php echo "{$field}[interval]"; ?>">Runs every:</label>
						<br/>
						<select name="<?php echo "{$field}[interval]"; ?>" id="<?php echo "{$field}[interval]"; ?>"
								style="width: 10em; vertical-align: middle">
							<?php
							global $WCLSI_JOB_INTERVALS;
							foreach ( $WCLSI_JOB_INTERVALS as $interval ) {
								if ( $selected_interval == $interval['value'] ) {
									echo "<option value=\"{$interval['value']}\" selected>{$interval['text']}</option>";
								} else {
									echo "<option value=\"{$interval['value']}\">{$interval['text']}</option>";
								}
							}
							?>
						</select>
					</fieldset>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_poller_setting_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			$field             = "{$this->plugin_id}{$this->id}_{$key}";
			$data              = wp_parse_args( $data );
			$is_enabled        = $this->settings[ $key ]['is_enabled'] ?? false;
			$selected_interval = $this->settings[ $key ]['interval'] ?? null;
			ob_start();
			?>
			<tr style="vertical-align: top;">
				<td scope="row" class="titledesc">
					<label for="<?php echo esc_attr( $field ); ?>"><?php echo wp_kses_post( $data['title'] ); ?></label>
				</td>
				<td>
					<?php echo $data['description']; ?>
				</td>
				<td class="forminp" style="width: 1em; vertical-align: top; padding: 15px 10px;">
					<fieldset>
						<label class="wclsi-switch">
							<input
									type="checkbox"
									id="<?php echo "{$field}[is_enabled]"; ?>"
									name="<?php echo "{$field}[is_enabled]"; ?>"
									value="true"
								<?php echo $is_enabled ? 'checked' : ''; ?>
							/>
							<span class="wclsi-slider round"></span>
						</label>
					</fieldset>
				</td>
				<td>
					<fieldset>
						<label for="<?php echo "{$field}[interval]"; ?>">Runs every:</label>
						<br/>
						<select name="<?php echo "{$field}[interval]"; ?>" id="<?php echo "{$field}[interval]"; ?>"
								style="width: 10em; vertical-align: middle">
							<?php
							global $WCLSI_JOB_INTERVALS;
							foreach ( $WCLSI_JOB_INTERVALS as $interval ) {
								if ( $selected_interval == $interval['value'] ) {
									echo "<option value=\"{$interval['value']}\" selected>{$interval['text']}</option>";
								} else {
									echo "<option value=\"{$interval['value']}\">{$interval['text']}</option>";
								}
							}
							?>
						</select>
					</fieldset>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_ls_to_wc_auto_load_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			$field             = "{$this->plugin_id}{$this->id}_{$key}";
			$data              = wp_parse_args( $data );
			$is_enabled        = $this->settings[ $key ]['is_enabled'] ?? false;
			$selected_interval = $this->settings[ $key ]['interval'] ?? null;
			ob_start();
			?>
			<tr style="vertical-align: top;">
				<td scope="row" class="titledesc">
					<label for="<?php echo esc_attr( $field ); ?>"><?php echo wp_kses_post( $data['title'] ); ?></label>
				</td>
				<td>
					<?php echo $data['description']; ?>
				</td>
				<td class="forminp" style="width: 1em; vertical-align: top; padding: 15px 10px;">
					<fieldset>
						<label class="wclsi-switch">
							<input
									type="checkbox"
									id="<?php echo "{$field}[is_enabled]"; ?>"
									name="<?php echo "{$field}[is_enabled]"; ?>"
									value="true"
								<?php echo $is_enabled ? 'checked' : ''; ?>
							/>
							<span class="wclsi-slider round"></span>
						</label>
					</fieldset>
				</td>
				<td>
					<fieldset>
						<div class="wclsi-background-job-setting">
							<label for="<?php echo "{$field}[interval]"; ?>">Runs every:</label>
							<br/>
							<select name="<?php echo "{$field}[interval]"; ?>" id="<?php echo "{$field}[interval]"; ?>"
									style="width: 10em;">
								<?php
								global $WCLSI_JOB_INTERVALS;
								foreach ( $WCLSI_JOB_INTERVALS as $interval ) {
									if ( $selected_interval == $interval['value'] ) {
										echo "<option value=\"{$interval['value']}\" selected>{$interval['text']}</option>";
									} else {
										echo "<option value=\"{$interval['value']}\">{$interval['text']}</option>";
									}
								}
								?>
							</select>
						</div>
						<div class="wclsi-background-job-setting">
							<label for="<?php echo "{$field}[autoload_after_action]"; ?>">New Lightspeed products should
								be:</label>
							<br/>
							<select name="<?php echo "{$field}[autoload_after_action]"; ?>" id="<?php echo $field; ?>"
									style="width: inherit;">
								<?php
								foreach ( WCLSI_Check_For_New_Prods::WCLSI_NEW_LS_PROD_BEHAVIOR as $opt_key => $description ) {
									if ( $this->settings[ $key ][ WCLSI_AUTOLOAD_AFTER_ACTION ] == $opt_key ) {
										echo "<option value=\"$opt_key\" selected>$description</option>";
									} else {
										echo "<option value=\"$opt_key\">$description</option>";
									}
								}
								?>
							</select>
						</div>
					</fieldset>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_ls_button_custom_html( $key, $data ) {

			$field                = "{$this->plugin_id}{$this->id}_{$key}";
			$data                 = wp_parse_args( $data );
			$woo_connect_full_url = add_query_arg(
				[
					'redirect' => rawurlencode( add_query_arg(
						[
							'page'      => 'wc-settings',
							'tab'       => 'integration',
							'section'   => 'lightspeed-integration',
							'reconnect' => ! empty( $this->token ) ? 'true' : 'false',
						],
						admin_url( 'admin.php' )
					) ),
					'nonce'    => wp_create_nonce( 'wclsi-auth' ),
				],
				$this->connect_url . '/lightspeed/auth'
			);

			$connect_text = ! empty( $this->token ) ? __( 'Reconnect to Lightspeed', 'woocommerce-lightspeed-pos' ) : __( 'Connect to Lightspeed', 'woocommerce-lightspeedpos' );

			ob_start();

			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<fieldset>
						<legend class="screen-reader-text">
							<span><?php echo wp_kses_post( $data['title'] ); ?></span>
						</legend>
						<?php echo $this->get_description_html( $data ); ?>
						<a href="<?php echo $woo_connect_full_url; ?>"
							class="wclsi-connect-ls"><?php echo $connect_text; ?></a>
					</fieldset>
				</td>
			</tr>
			<?php

			return ob_get_clean();
		}

		function generate_wclsi_sku_strategy_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}
			$field           = "{$this->plugin_id}{$this->id}_{$key}";
			$data            = wp_parse_args( $data );
			$selected_option = isset( $this->settings[ $key ] ) ? $this->settings[ $key ] : 'plugin_default';
			global $WCLSI_SKU_STRATEGIES;
			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<div style="max-width: 400px;">
						<fieldset>
							<legend class="screen-reader-text">
								<span><?php echo wp_kses_post( $data['title'] ); ?></span>
							</legend>
							<?php echo $this->get_description_html( $data ); ?>
						</fieldset>
						<select name="<?php echo $field; ?>" id="<?php echo $field; ?>" style="width: inherit;">
							<?php
							foreach ( $WCLSI_SKU_STRATEGIES as $option ) {
								$selected = $selected_option === $option['value'] ? 'selected' : '';
								echo "<option value='{$option['value']}' {$selected}>{$option['text']}</option>";
							}
							?>
						</select>
					</div>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_image_compression_strategy_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}
			$field           = "{$this->plugin_id}{$this->id}_{$key}";
			$data            = wp_parse_args( $data );
			$selected_option = isset( $this->settings[ $key ] ) ? $this->settings[ $key ] : 'plugin_default';
			global $WCLSI_IMAGE_COMPRESSION_STRATEGIES;
			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<div style="max-width: 400px;">
						<fieldset>
							<legend class="screen-reader-text">
								<span><?php echo wp_kses_post( $data['title'] ); ?></span>
							</legend>
							<?php echo $this->get_description_html( $data ); ?>
						</fieldset>
						<select name="<?php echo $field; ?>" id="<?php echo $field; ?>" style="width: inherit;">
							<?php
							foreach ( $WCLSI_IMAGE_COMPRESSION_STRATEGIES as $option ) {
								$selected = $selected_option === $option['value'] ? 'selected' : '';
								echo "<option value='{$option['value']}' {$selected}>{$option['text']}</option>";
							}
							?>
						</select>
					</div>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_handle_archived_products_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}
			$field           = "{$this->plugin_id}{$this->id}_{$key}";
			$data            = wp_parse_args( $data );
			$selected_option = isset( $this->settings[ $key ] ) ? $this->settings[ $key ] : 'plugin_default';
			global $WCLSI_HANDLE_ARCHIVED_PRODUCT_STRATEGIES;
			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<div style="max-width: 400px;">
						<fieldset>
							<legend class="screen-reader-text">
								<span><?php echo wp_kses_post( $data['title'] ); ?></span>
							</legend>
							<?php echo $this->get_description_html( $data ); ?>
						</fieldset>
						<select name="<?php echo $field; ?>" id="<?php echo $field; ?>" style="width: inherit;">
							<?php
							foreach ( $WCLSI_HANDLE_ARCHIVED_PRODUCT_STRATEGIES as $option ) {
								$selected = $selected_option === $option['value'] ? 'selected' : '';
								echo "<option value='{$option['value']}' {$selected}>{$option['text']}</option>";
							}
							?>
						</select>
					</div>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_import_status_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}
			$field    = "{$this->plugin_id}{$this->id}_{$key}";
			$data     = wp_parse_args( $data );
			$selected = isset( $this->settings[ $key ] ) ? $this->settings[ $key ] : 'draft';
			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<div style="max-width: 400px;">
						<fieldset>
							<legend class="screen-reader-text">
								<span><?php echo wp_kses_post( $data['title'] ); ?></span>
							</legend>
							<?php echo $this->get_description_html( $data ); ?>
						</fieldset>
						<select name="<?php echo $field; ?>" id="<?php echo $field; ?>" style="width: inherit;">
							<option value="draft"
								<?php
								if ( $selected == 'draft' || empty( $selected ) ) {
									echo 'selected';
								}
								?>
								>
								Draft (default)
							</option>
							<option value="publish" <?php
							if ( $selected == 'publish' ) {
								echo 'selected';
							}
							?>
							>
								Publish
							</option>
						</select>
						<?php if ( $selected == 'publish' ) { ?>
							<div>
								<p class="notice notice-warning">
									<?php
									echo __(
										'Warning: this will make your products live and visible on your store catalog upon import.',
										'woocommerce-lightspeed-pos'
									)
									?>
								</p>
							</div>
						<?php } ?>
					</div>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_wc_selective_sync_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}
			global $WC_PROD_SELECTIVE_SYNC_PROPERTIES;
			$field           = "{$this->plugin_id}{$this->id}_{$key}";
			$data            = wp_parse_args( $data );
			$sync_properties = $this->settings['wclsi_wc_selective_sync'];
			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<div style="max-width: 400px;" class="wclsi-selective-sync-checkboxes">
						<fieldset>
							<legend class="screen-reader-text">
								<span><?php echo wp_kses_post( $data['title'] ); ?></span>
							</legend>
							<?php echo $this->get_description_html( $data ); ?>
							<p id="wclsi_select_all_prod_sync_properties">Select all</p>
							<div id="wc_prod_selective_sync_properties">
								<?php foreach ( $WC_PROD_SELECTIVE_SYNC_PROPERTIES as $key => $property ) : ?>
									<?php if ( $property == 'Short Description' ) : ?>
										<p id="wclsi-legacy-webstore-header">Legacy Web Store fields: </p>
									<?php endif; ?>
									<input
											type="checkbox"
											name="<?php echo "{$field}[$key]"; ?>"
											id="<?php echo "{$field}[$key]"; ?>"
											value="true"
										<?php
										if ( isset( $sync_properties[ $key ] ) && $sync_properties[ $key ] == 'true' ) {
											echo 'checked';
										}
										?>
									/>
									<label for="<?php echo "{$field}[$key]"; ?>"><?php echo $property; ?></label>
									<br/>
								<?php endforeach; ?>
							</div>
						</fieldset>
						<div id="wclsi-selective-sync-gradient"></div>
					</div>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_ls_selective_sync_html( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}
			global $LS_PROD_SELECTIVE_SYNC_PROPERTIES;
			$field = "{$this->plugin_id}{$this->id}_{$key}";
			$data  = wp_parse_args( $data );
			if ( isset( $this->settings[ WCLSI_LS_SELECTIVE_SYNC ] ) ) {
				$sync_properties = $this->settings[ WCLSI_LS_SELECTIVE_SYNC ];
			} else {
				$sync_properties = [];
			}

			ob_start();
			?>
			<tr>
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp">
					<div style="max-width: 400px;" class="wclsi-selective-sync-checkboxes">
						<fieldset>
							<legend class="screen-reader-text">
								<span><?php echo wp_kses_post( $data['title'] ); ?></span>
							</legend>
							<?php echo $this->get_description_html( $data ); ?>
							<p id="wclsi_select_all_ls_prod_sync_properties">Select all</p>
							<div id="ls_prod_selective_sync_properties">
								<?php foreach ( $LS_PROD_SELECTIVE_SYNC_PROPERTIES as $key => $property ) : ?>
									<?php if ( $property == 'Short Description' ) : ?>
										<p id="wclsi-legacy-webstore-header"> Legacy Web Store fields: </p>
									<?php endif; ?>
									<input
											type="checkbox"
											name="<?php echo "{$field}[$key]"; ?>"
											id="<?php echo "{$field}[$key]"; ?>"
											value="true"
										<?php echo isset( $sync_properties[ $key ] ) && $sync_properties[ $key ] == 'true' ? 'checked' : ''; ?>
									/>
									<label for="<?php echo "{$field}[$key]"; ?>"><?php echo $property; ?></label>
									<br/>
								<?php endforeach; ?>
							</div>
						</fieldset>
					</div>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function generate_wclsi_empty_on_zero_sale_price_html( $key, $data ) {
			return $this->single_checkbox( $key, $data );
		}

		function generate_wclsi_ignore_archived_ls_prods_html( $key, $data ) {
			return $this->single_checkbox( $key, $data );
		}

		function generate_wclsi_prune_deleted_variations_html( $key, $data ) {
			return $this->single_checkbox( $key, $data );
		}

		function tooltip_th( $field, $data ) {
			ob_start();
			?>
			<th scope="row" class="titledesc">
				<label for="<?php echo esc_attr( $field ); ?>"><?php echo wp_kses_post( $data['title'] ); ?></label>
				<?php echo $this->get_tooltip_html( $data ); ?>
			</th>
			<?php
			return ob_get_clean();
		}

		function single_checkbox( $key, $data ) {
			if ( ! self::wclsi_initialized() ) {
				return '';
			}

			$field = "{$this->plugin_id}{$this->id}_{$key}";
			$data  = wp_parse_args( $data );

			$setting = isset( $this->settings[ $key ] ) ? $this->settings[ $key ] : false;
			$checked = ! empty( $setting ) ? 'checked' : '';

			ob_start();
			?>
			<tr style="vertical-align: top;">
				<?php echo $this->tooltip_th( $field, $data ); ?>
				<td class="forminp" style="vertical-align: top; padding: 15px 10px;">
					<fieldset>
						<?php echo $this->get_description_html( $data ); ?>
						<label>
							<input
									type="checkbox"
									name="<?php echo $field; ?>"
									id="<?php echo $field; ?>"
									value="true"
								<?php echo $checked; ?>
							/>
						</label>
						<br/>
					</fieldset>
				</td>
			</tr>
			<?php
			return ob_get_clean();
		}

		function validate_wclsi_poller_setting_field( $key ) {
			$setting_key  = "{$this->plugin_id}{$this->id}_{$key}";
			$new_interval = $_POST[ $setting_key ]['interval'] ?? false;
			if ( ! $this->validate_job_interval( $new_interval ) ) {
				$this->add_error( "Invalid interval value: {$new_interval}" );

				return $this->settings[ $key ];
			}

			// unschedule the next job so we create a new one with the proper interval
			$current_interval = $this->settings[ WCLSI_POLLER_SETTING ]['interval'] ?? null;
			if ( $current_interval !== $new_interval ) {
				as_unschedule_all_actions( 'wclsi_poll' );
			}

			return $_POST[ $setting_key ];
		}

		function validate_wclsi_ls_to_wc_auto_load_field( $key ) {
			$setting_key  = "{$this->plugin_id}{$this->id}_{$key}";
			$new_interval = $_POST[ $setting_key ]['interval'] ?? false;
			if ( ! $this->validate_job_interval( $new_interval ) ) {
				$this->add_error( "Invalid interval value: {$new_interval}" );

				return $this->settings[ $key ];
			}

			// unschedule the next job so we create a new one with the proper interval
			$current_interval = $this->settings[ WCLSI_LS_TO_WC_AUTOLOAD ]['interval'] ?? null;
			if ( $current_interval !== $new_interval ) {
				as_unschedule_all_actions( 'check_for_new_ls_prods' );
			}

			return $_POST[ $setting_key ];
		}

		function validate_wclsi_import_on_auto_load_field( $key ) {
			return $_POST[ "{$this->plugin_id}{$this->id}_{$key}" ] ?? $_POST[ "{$this->plugin_id}{$this->id}_{$key}" ] ?? null;
		}

		function validate_wclsi_wc_selective_sync_field( $key ) {
			return $_POST[ "{$this->plugin_id}{$this->id}_{$key}" ] ?? $_POST[ "{$this->plugin_id}{$this->id}_{$key}" ] ?? null;
		}

		function validate_wclsi_ls_selective_sync_field( $key ) {
			if ( ! isset( $_POST[ "{$this->plugin_id}{$this->id}_{$key}" ] ) ) {
				return null;
			}

			if ( 'wclsi_ls_selective_sync' === $key && key_exists( 'sale_price', $_POST[ "{$this->plugin_id}{$this->id}_{$key}" ] ) ) {
				if ( ! $this->sale_price_level_enabled() ) {
					unset( $_POST[ "{$this->plugin_id}{$this->id}_{$key}" ]['sale_price'] );
				}
			}

			return $_POST[ "{$this->plugin_id}{$this->id}_{$key}" ];
		}

		function validate_wclsi_poll_prod_attribute_sets_field( $key ) {
			$setting_key  = "{$this->plugin_id}{$this->id}_{$key}";
			$new_interval = $_POST[ $setting_key ]['interval'] ?? false;
			if ( ! $this->validate_job_interval( $new_interval ) ) {
				$this->add_error( "Invalid interval value: {$new_interval}" );

				return $this->settings[ $key ];
			}

			// unschedule the next job so we create a new one with the proper interval
			$current_interval = $this->settings[ WCLSI_ATTR_SETS_POLLER_SETTINGS ]['interval'] ?? null;
			if ( $current_interval !== $new_interval ) {
				as_unschedule_all_actions( 'wclsi_poll_ls_attribute_sets' );
			}

			return $_POST[ $setting_key ];
		}

		private function validate_job_interval( $interval ): bool {
			global $WCLSI_JOB_INTERVALS;

			return in_array( $interval, array_keys( $WCLSI_JOB_INTERVALS ) );
		}

		function sale_price_level_enabled() {
			global $WCLSI;
			$price_level_endpt = "Account/$WCLSI->ls_account_id/PriceLevel/";
			$results           = $WCLSI->make_api_call( $price_level_endpt, 'Read', [ 'name' => 'Sale' ] );
			if ( ! is_wp_error( $results ) ) {
				if ( $results->{'@attributes'}->count > 0 ) {
					return true;
				} else {
					$this->init_errors[] =
						'Could not save "Sale Price" for Lightspeed Selective Sync. ' .
						'Please set up a "Sale" PriceLevel in your Lightspeed store first! ' .
						'Please refer to the <a href="https://docs.woocommerce.com/document/woocommerce-lightspeed-pos/">documentation</a> for more info.';
				}
			} else {
				$this->init_errors[] =
					'Could not save "Sale Price" for Lightspeed Selective Sync. ' .
					'It could not be determined if your Lightspeed store has a "Sale" price level set up. ' .
					'Please refer to the <a href="https://docs.woocommerce.com/document/woocommerce-lightspeed-pos/">documentation</a> for more info.';
			}

			return false;
		}

		function clear_table_data_ajax() {
			wclsi_verify_nonce();

			global $wpdb;
			$wpdb->show_errors();

			$this->clear_table_data();

			if ( ! empty( $wpdb->last_error ) ) {
				echo json_encode( [ 'errors' => $wpdb->last_error ] );
			} else {
				echo json_encode( [ 'success' => true ] );
			}

			exit;
		}

		private function clear_table_data() {
			if ( ! is_admin() ) {
				return;
			}

			global $wpdb, $WCLSI_ITEM_TABLE, $WCLSI_ITEM_E_COMMERCE_TABLE, $WCLSI_ITEM_IMAGES_TABLE,
					$WCLSI_ITEM_PRICES_TABLE, $WCLSI_ITEM_SHOP_TABLE, $WCLSI_ITEM_ATTRIBUTE_SETS_TABLE;

			$tables = [
				$WCLSI_ITEM_TABLE,
				$WCLSI_ITEM_SHOP_TABLE,
				$WCLSI_ITEM_PRICES_TABLE,
				$WCLSI_ITEM_IMAGES_TABLE,
				$WCLSI_ITEM_E_COMMERCE_TABLE,
				$WCLSI_ITEM_ATTRIBUTE_SETS_TABLE,
			];

			foreach ( $tables as $table ) {
				$wpdb->query( "TRUNCATE $table;" ); // phpcs:ignore
			}
		}

		function display_init_errors() {

			$displayed = wp_cache_get( 'wclsi_init_errors_displayed', WCLSI_CACHE );

			if ( ! $displayed && is_array( $this->init_errors ) && ! empty( $this->init_errors ) ) :

				foreach ( $this->init_errors as $key => $error ) :

					?>
					<div class="error is-dismissible'">
						<p><?php echo $error; ?></p>
					</div>
					<?php

				endforeach;

				wp_cache_set( 'wclsi_init_errors_displayed', true, WCLSI_CACHE );

			endif;
		}

		function check_for_lightspeed_token() {
			if ( ! empty( $_GET['wclsi_access_token'] ) ) {

				// ensure the access token data is from our original redirect to the connect app
				if ( ! wp_verify_nonce( wp_unslash( $_GET['wclsi_nonce'] ) ?? '', 'wclsi-auth' ) ) {
					wp_die( __( 'Unable to verify nonce, please refresh the page and try connecting to Lightspeed again.', 'woocommerce-lightspeed-pos' ) );
				}

				$token         = wc_clean( wp_unslash( $_GET['wclsi_access_token'] ) );
				$refresh_token = wc_clean( wp_unslash( $_GET['wclsi_refresh_token'] ) ?? '' );
				$expires_in    = intval( wp_unslash( $_GET['wclsi_expires_in'] ) ?? '' );

				// Setup a separate MerchantOS object since we may not have the account ID yet..
				$ls_api = new WP_MOSAPICall( '', $token );

				$this->init_wclsi_settings_with_token( $ls_api, $token, $refresh_token, $expires_in );
			}
		}

		function init_wclsi_settings_with_token( $ls_api, $token, $refresh_token = null, $expires_in = null ) {

			// Save account settings
			try {
				$account = $ls_api->makeAPICall( 'Account', 'Read' );
			} catch ( Exception $e ) {
				$this->init_errors[] = $e->getMessage();

				return;
			}

			if ( isset( $account->httpCode ) && $account->httpCode != '200' ) {
				$this->init_errors[] =
					sprintf(
						'%s %d %s %s %s %s',
						'Lightspeed API Error - ',
						$account->httpCode,
						$account->message,
						'Read',
						' - Payload: ',
						'/Account'
					);

				return;
			} else {
				$this->check_for_ls_account( $account, $token );
			}

			// Get Shop info
			try {
				$shop = $ls_api->makeAPICall( "Account/{$this->ls_account_id}/Shop/", 'Read' );
			} catch ( Exception $e ) {
				$this->init_errors[] = $e->getMessage();

				return;
			}

			if ( isset( $shop->httpCode ) && $shop->httpCode != '200' ) {
				$this->init_errors[] =
					sprintf(
						'%s %d %s %s %s %s',
						'Lightspeed API Error - ',
						$shop->httpCode,
						$shop->message,
						'Read',
						' - Payload: ',
						'/Shop'
					);

				return;
			} else {
				$this->check_for_ls_shop_data( $shop );
			}

			// Plugin is no longer clean, display errors if certain options are not available
			update_option( 'wclsi_initialized', true );
			update_option( 'wclsi_oauth_token', $token );
			update_option( 'wclsi_refresh_token', $refresh_token );

			if ( $expires_in > 0 ) {
				$wclsi_expires_in = date( 'Y-m-d H:i:s', time() + $expires_in );
				update_option( 'wclsi_expires_in', $wclsi_expires_in );
			}
		}

		public static function wclsi_initialized() {
			$wclsi_account_id = get_option( 'wclsi_account_id' );

			return false !== $wclsi_account_id;
		}

		/**
		 * @param              $controlname
		 * @param              $action
		 * @param string $query_str
		 * @param array $data
		 * @param null $unique_id
		 * @param Closure|null $callback
		 *
		 * @return array|mixed|object|WP_Error
		 */
		function make_api_call(
			$controlname,
			$action,
			$query_str = '',
			$data = [],
			$unique_id = null,
			Closure $callback = null
		) {

			// Stub out API calls if we are testing
			if ( 'true' === getenv( 'WCLSI_TEST' ) ) {
				return false;
			}

			try {
				$controlname = apply_filters( 'wclsi_api_call_controlname', $controlname );
				$action      = apply_filters( 'wclsi_api_call_action', $action );
				$query_str   = apply_filters( 'wclsi_api_call_query_str', $query_str );
				$data        = apply_filters( 'wclsi_api_call_data', $data );
				$unique_id   = apply_filters( 'wclsi_api_call_unique_id', $unique_id );

				$result = $this->MOSAPI->makeAPICall( $controlname, $action, $unique_id, $data, $query_str, $callback );

				if ( isset( $result->httpCode ) ) {

					// Handle refresh tokens
					if ( $result->httpCode == '401' ) {
						$token_result = $this->refresh_access_token();

						if ( ! is_wp_error( $token_result ) ) {
							$this->MOSAPI = new WP_MOSAPICall( $this->ls_account_id, $token_result );

							// Try the request again once the token has been refreshed
							$result = $this->MOSAPI->makeAPICall(
								$controlname, $action, $unique_id, $data, $query_str, $callback
							);
						} else {
							if ( is_admin() ) {
								add_settings_error(
									'wclsi_settings',
									'wclsi_bad_refresh_token_attempt',
									'Error: ' . $token_result->get_error_message()
								);
							}

							return new WP_Error(
								'wclsi_bad_api_call',
								$token_result->get_error_message()
							);
						}
					}

					// Handle other types of status codes other than 200
					// We need another isset() check here in case there was a token refresh
					if ( isset( $result->httpCode ) && $result->httpCode != '200' ) {

						$error_msg = sprintf(
							'%s %d %s %s %s %s %s %s %s',
							__( 'Lightspeed API Error - ', 'woocommerce-lightspeed-pos' ),
							$result->httpCode,
							$result->message,
							$action,
							__( ' - Payload: ', 'woocommerce-lightspeed-pos' ),
							$controlname,
							$unique_id,
							print_r( $query_str, true ),
							print_r( $data, true )
						);

						if ( is_admin() && function_exists( 'add_settings_error' ) ) {
							add_settings_error(
								'wclsi_settings',
								'wclsi_bad_api_call',
								$error_msg,
								'error'
							);
						}

						// WCLSI_Logger::log_error("Lightspeed API call failed", $result, true);
						return new WP_Error( 'wclsi_bad_api_call', $error_msg );
					}
				}

				return $result;
			} catch ( Exception $e ) {
				if ( is_admin() && function_exists( 'add_settings_error' ) ) {
					add_settings_error(
						'wclsi_settings',
						'wclsi_bad_api_call',
						__( $e->getMessage(), 'woocommerce-lightspeed-pos' )
					);
				}

				WCLSI_Logger::log_error( "Lightspeed API call failed: {$e->getMessage()}", $e, true );

				return new WP_Error( 'wclsi_bad_api_call', __( $e->getMessage(), 'woocommerce-lightspeed-pos' ) );
			}
		}

		/**
		 * Set up default settings
		 */
		public function init_settings() {
			parent::init_settings();
			$this->seed_selective_sync();
			$this->seed_autoload_job_settings();
			$this->seed_wclsi_poller_job_settings();
			$this->seed_wclsi_attr_sets_poller_job_settings();
			$this->seed_ignore_archived_ls_prods();
			$this->seed_prune_deleted_variations();
			$this->seed_wclsi_empty_on_zero_sale_price();
			$this->seed_wclsi_image_compression_setting();
			$this->seed_wclsi_handle_archived_products_setting();
			$this->seed_wclsi_sales_sync_setting();
		}

		/*******************
		 * Private Methods *
		 *******************/
		private function seed_wclsi_handle_archived_products_setting() {
			if ( ! array_key_exists( 'wclsi_handle_archived_products_strategy', $this->settings ) ) {
				$this->settings['wclsi_handle_archived_products_strategy'] = 'do nothing';
			}
		}

		private function seed_wclsi_image_compression_setting() {
			if ( ! array_key_exists( 'wclsi_image_compression_strategy', $this->settings ) ) {
				$this->settings['wclsi_image_compression_strategy'] = 'q_auto:eco';
			}
		}

		private function seed_wclsi_sales_sync_setting() {
			if ( ! array_key_exists( 'wclsi_sales_sync', $this->settings ) ) {
				$this->settings['wclsi_sales_sync'] = 'false';
			}
		}

		private function seed_wclsi_empty_on_zero_sale_price() {
			if ( ! array_key_exists( 'wclsi_empty_on_zero_sale_price', $this->settings ) ) {
				$this->settings['wclsi_empty_on_zero_sale_price'] = 'true';
			}
		}

		private function seed_wclsi_poller_job_settings() {
			if ( ! array_key_exists( WCLSI_POLLER_SETTING, $this->settings ) ) {
				$this->settings[ WCLSI_POLLER_SETTING ]               = [];
				$this->settings[ WCLSI_POLLER_SETTING ]['is_enabled'] = 'true';
				$this->settings[ WCLSI_POLLER_SETTING ]['interval']   = 60;
			}
		}

		private function seed_wclsi_attr_sets_poller_job_settings() {
			if ( ! array_key_exists( WCLSI_ATTR_SETS_POLLER_SETTINGS, $this->settings ) ) {
				$this->settings[ WCLSI_ATTR_SETS_POLLER_SETTINGS ]               = [];
				$this->settings[ WCLSI_ATTR_SETS_POLLER_SETTINGS ]['is_enabled'] = 'true';
				$this->settings[ WCLSI_ATTR_SETS_POLLER_SETTINGS ]['interval']   = 60;
			}
		}

		private function seed_autoload_job_settings() {
			if ( ! array_key_exists( WCLSI_LS_TO_WC_AUTOLOAD, $this->settings ) ) {
				$this->settings[ WCLSI_LS_TO_WC_AUTOLOAD ]               = [];
				$this->settings[ WCLSI_LS_TO_WC_AUTOLOAD ]['is_enabled'] = 'true';
				$this->settings[ WCLSI_LS_TO_WC_AUTOLOAD ]['interval']   = 60;

				if ( ! array_key_exists( WCLSI_AUTOLOAD_AFTER_ACTION, $this->settings[ WCLSI_LS_TO_WC_AUTOLOAD ] ) ) {
					$this->settings[ WCLSI_LS_TO_WC_AUTOLOAD ][ WCLSI_AUTOLOAD_AFTER_ACTION ] = WCLSI_Check_For_New_Prods::LOAD_INTO_IMPORTER;
				}
			}
		}

		private function seed_prune_deleted_variations() {
			if ( ! array_key_exists( WCLSI_PRUNE_DELETED_VARIATIONS, $this->settings ) ) {
				$this->settings[ WCLSI_PRUNE_DELETED_VARIATIONS ] = 'true';
			}
		}

		private function seed_ignore_archived_ls_prods() {
			if ( ! array_key_exists( WCLSI_IGNORE_ARCHIVED_LS_PRODS, $this->settings ) ) {
				$this->settings[ WCLSI_IGNORE_ARCHIVED_LS_PRODS ] = 'true';
			}
		}

		private function seed_selective_sync() {

			if ( ! array_key_exists( WCLSI_WC_SELECTIVE_SYNC, $this->settings ) ) {
				global $WC_PROD_SELECTIVE_SYNC_PROPERTIES;

				// default all selective sync properties to true
				$selective_sync_default    = [];
				$selective_ls_sync_default = [];

				foreach ( $WC_PROD_SELECTIVE_SYNC_PROPERTIES as $key => $property ) {

					// We're not sure if the Sale Price Level has been setup, leave it out
					if ( $key === 'sale_price' ) {
						continue;
					}

					$selective_sync_default[ $key ] = 'true';
				}

				$this->settings['wclsi_wc_selective_sync'] = $selective_sync_default;
			}

			if ( ! array_key_exists( WCLSI_LS_SELECTIVE_SYNC, $this->settings ) ) {

				// default ony the stock_quantity & stock_quantity_checkout sync property to true
				$selective_ls_sync_default['stock_quantity'] = 'true';
				$this->settings[ WCLSI_LS_SELECTIVE_SYNC ]   = $selective_ls_sync_default;
			}
		}

		private function set_header_info() {
			$this->id                 = 'lightspeed-integration';
			$this->method_title       = __( 'Lightspeed Retail POS', 'woocommerce-lightspeed-pos' );
			$this->method_description =
				'<div id="wclsi-js-notifications"></div>' .
				'<p>' .
				__( 'Import Lightspeed Cloud data to your WooCommerce instance', 'woocommerce-lightspeed-pos' ) . ' | ' .
				'<a href="' . WCLSI_ADMIN_URL . '">Importer</a>' . ' | ' .
				'<a href="' . WCLSI_DOCS_URL . '" target="_blank">Documentation</a>' . ' | ' .
				'<a href="https://dzv365zjfbd8v.cloudfront.net/changelogs/woocommerce-lightspeed-pos/changelog.txt">v' . WCLSI_VERSION . '</a>' .
				'</p>';
		}

		private function init_store_data() {
			$shop_data = get_option( 'wclsi_shop_data' );
			if ( ! empty( $shop_data ) ) {
				$this->store_timezone = $shop_data['store_timezone'];
				$this->store_name     = $shop_data['store_name'];
			}
		}

		private function render_api_success_message() {
			if ( wclsi_oauth_enabled() && ! empty( $this->store_name ) && ! empty( $this->store_timezone ) ) {
				printf(
					'<p id="wclsi-api-status" style="%s">%s<strong>%s</strong>, %s<strong>%s</strong></p>',
					'color: green;',
					__( ' API Settings successfully initialized! Store name(s): ', 'woocommerce-lightspeed-pos' ),
					$this->store_name,
					__( 'Time Zone: ', 'woocommerce-lightspeed-pos' ),
					$this->store_timezone
				);
			}
		}

		private function check_for_ls_account( $account_response, $token ) {
			if ( isset( $account_response->Account->accountID ) ) {

				// Save the token
				$this->token    = $token;
				$wclsi_settings = get_option( $this->get_settings_option_key() );

				// Delete the deprecated API Key if it exists
				if ( isset( $wclsi_settings['api_key'] ) ) {
					unset( $wclsi_settings['api_key'] );
				}

				update_option( $this->get_settings_option_key(), $wclsi_settings );

				$this->ls_account_id = $account_response->Account->accountID;

				// Save the account ID
				update_option( 'wclsi_account_id', $account_response->Account->accountID );
			} else {
				$this->init_errors[] = __( 'Could not find an account associated with this API key.', 'woocommerce-lightspeed-pos' );

				return;
			}
		}

		private function check_for_ls_shop_data( $shop_response ) {

			// Note: in the case of multiple shops, the time-zone will default to the first store
			if ( ! empty( $shop_response ) ) {

				// Add it to the settings
				$shop_data                   = [];
				$shop_data['store_timezone'] = is_array( $shop_response->Shop ) ? $shop_response->Shop[0]->timeZone : $shop_response->Shop->timeZone;
				$shop_data['store_name']     = is_array( $shop_response->Shop ) ? 'Multiple' : $shop_response->Shop->name;
				$shop_data['ls_store_data']  = $shop_response;

				if ( is_object( $shop_response->Shop ) ) {
					$wc_options = get_option( $this->get_settings_option_key() );
					$wc_options[ WCLSI_INVENTORY_SHOP_ID ] = $shop_response->Shop->shopID;
					update_option( $this->get_settings_option_key(), $wc_options );
				}

				update_option( 'wclsi_shop_data', $shop_data );
			}
		}

		private function refresh_access_token() {
			$connect_url = add_query_arg(
				[
					'site_url'      => site_url(),
					'refresh_token' => get_option( 'wclsi_refresh_token' ),
				],
				$this->connect_url . '/lightspeed/refresh-access'
			);

			$result = wp_remote_post( $connect_url );

			if ( is_wp_error( $result ) ) {
				return $result;
			}

			if ( $result['response']['code'] == 200 ) {
				$response = json_decode( $result['body'] );

				if ( isset( $response->access_token ) ) {
					update_option( 'wclsi_oauth_token', $response->access_token );

					return $response->access_token;
				} elseif ( isset( $response->error ) && isset( $response->reason ) ) {
					return new WP_Error( 'could_not_retrieve_access_token', $response->reason, $result );
				}
			}

			return new WP_Error(
				'could_not_retrieve_access_token',
				'Something went wrong with refreshing the access token. Please contact support!',
				$result
			);
		}
	}

	global $WCLSI;
	$WCLSI = new WCLSI_Settings();
endif;
// phpcs:enable
